#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition to achieve a Swap gate.

Decomposes a Swap gate using 3 CNOT gates, where the one in the middle
features as many control qubits as the Swap gate has control qubits.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Uncompute, Control
from projectq.ops import Swap, CNOT


def _decompose_swap(cmd):
    """ Decompose (controlled) swap gates. """
    ctrl = cmd.control_qubits
    eng = cmd.engine
    with Compute(eng):
        CNOT | (cmd.qubits[0], cmd.qubits[1])
    with Control(eng, ctrl):
        CNOT | (cmd.qubits[1], cmd.qubits[0])
    Uncompute(eng)


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Swap.__class__, _decompose_swap)
]



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition to for a CNOT gate in terms of CZ and Hadamard.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, get_control_count, Uncompute
from projectq.ops import CZ, H, X


def _decompose_cnot(cmd):
    """ Decompose CNOT gates. """
    ctrl = cmd.control_qubits
    eng = cmd.engine
    with Compute(eng):
        H | cmd.qubits[0]
    CZ | (ctrl[0], cmd.qubits[0][0])
    Uncompute(eng)


def _recognize_cnot(cmd):
    return get_control_count(cmd) == 1


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(X.__class__, _decompose_cnot, _recognize_cnot)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for the Toffoli gate.

Decomposes the Toffoli gate using Hadamard, T, Tdag, and CNOT gates.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import NOT, CNOT, T, Tdag, H


def _decompose_toffoli(cmd):
    """ Decompose the Toffoli gate into CNOT, H, T, and Tdagger gates. """
    ctrl = cmd.control_qubits

    target = cmd.qubits[0]
    c1 = ctrl[0]
    c2 = ctrl[1]

    H | target
    CNOT | (c1, target)
    T | c1
    Tdag | target
    CNOT | (c2, target)
    CNOT | (c2, c1)
    Tdag | c1
    T | target
    CNOT | (c2, c1)
    CNOT | (c1, target)
    Tdag | target
    CNOT | (c2, target)
    T | target
    T | c2
    H | target


def _recognize_toffoli(cmd):
    """ Recognize the Toffoli gate. """
    return get_control_count(cmd) == 2


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(NOT.__class__, _decompose_toffoli, _recognize_toffoli)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
""" Back-end to run quantum program on IBM's Quantum Experience."""
import math
import random

from projectq.cengines import BasicEngine
from projectq.meta import get_control_count, LogicalQubitIDTag
from projectq.ops import (NOT, H, Rx, Ry, Rz, Measure, Allocate, Deallocate,
                          Barrier, FlushGate)

from ._ibm_http_client import send, retrieve


class IBMBackend(BasicEngine):
    """
    The IBM Backend class, which stores the circuit, transforms it to JSON,
    and sends the circuit through the IBM API.
    """
    def __init__(self,
                 use_hardware=False,
                 num_runs=1024,
                 verbose=False,
                 token='',
                 device='ibmq_essex',
                 num_retries=3000,
                 interval=1,
                 retrieve_execution=None):
        """
        Initialize the Backend object.

        Args:
            use_hardware (bool): If True, the code is run on the IBM quantum
                chip (instead of using the IBM simulator)
            num_runs (int): Number of runs to collect statistics.
                (default is 1024)
            verbose (bool): If True, statistics are printed, in addition to
                the measurement result being registered (at the end of the
                circuit).
            token (str): IBM quantum experience user password.
            device (str): name of the IBM device to use. ibmq_essex By default
            num_retries (int): Number of times to retry to obtain
                results from the IBM API. (default is 3000)
            interval (float, int): Number of seconds between successive
                attempts to obtain results from the IBM API.
                (default is 1)
            retrieve_execution (int): Job ID to retrieve instead of re-
                running the circuit (e.g., if previous run timed out).
        """
        BasicEngine.__init__(self)
        self._reset()
        if use_hardware:
            self.device = device
        else:
            self.device = 'ibmq_qasm_simulator'
        self._num_runs = num_runs
        self._verbose = verbose
        self._token = token
        self._num_retries = num_retries
        self._interval = interval
        self._probabilities = dict()
        self.qasm = ""
        self._json = []
        self._measured_ids = []
        self._allocated_qubits = set()
        self._retrieve_execution = retrieve_execution

    def is_available(self, cmd):
        """
        Return true if the command can be executed.

        The IBM quantum chip can only do U1,U2,U3,barriers, and CX / CNOT.
        Conversion implemented for Rotation gates and H gates.

        Args:
            cmd (Command): Command for which to check availability
        """
        g = cmd.gate
        if g == NOT and get_control_count(cmd) == 1:
            return True
        if get_control_count(cmd) == 0:
            if g == H:
                return True
            if isinstance(g, (Rx, Ry, Rz)):
                return True
        if g in (Measure, Allocate, Deallocate, Barrier):
            return True
        return False

    def get_qasm(self):
        """ Return the QASM representation of the circuit sent to the backend.
        Should be called AFTER calling the ibm device """
        return self.qasm

    def _reset(self):
        """ Reset all temporary variables (after flush gate). """
        self._clear = True
        self._measured_ids = []

    def _store(self, cmd):
        """
        Temporarily store the command cmd.

        Translates the command and stores it in a local variable (self._cmds).

        Args:
            cmd: Command to store
        """
        if self._clear:
            self._probabilities = dict()
            self._clear = False
            self.qasm = ""
            self._json = []
            self._allocated_qubits = set()

        gate = cmd.gate
        if gate == Allocate:
            self._allocated_qubits.add(cmd.qubits[0][0].id)
            return

        if gate == Deallocate:
            return

        if gate == Measure:
            assert len(cmd.qubits) == 1 and len(cmd.qubits[0]) == 1
            logical_id = None
            for t in cmd.tags:
                if isinstance(t, LogicalQubitIDTag):
                    logical_id = t.logical_qubit_id
                    break
            assert logical_id is not None
            self._measured_ids += [logical_id]
        elif gate == NOT and get_control_count(cmd) == 1:
            ctrl_pos = cmd.control_qubits[0].id
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\ncx q[{}], q[{}];".format(ctrl_pos, qb_pos)
            self._json.append({'qubits': [ctrl_pos, qb_pos], 'name': 'cx'})
        elif gate == Barrier:
            qb_pos = [qb.id for qr in cmd.qubits for qb in qr]
            self.qasm += "\nbarrier "
            qb_str = ""
            for pos in qb_pos:
                qb_str += "q[{}], ".format(pos)
            self.qasm += qb_str[:-2] + ";"
            self._json.append({'qubits': qb_pos, 'name': 'barrier'})
        elif isinstance(gate, (Rx, Ry, Rz)):
            assert get_control_count(cmd) == 0
            qb_pos = cmd.qubits[0][0].id
            u_strs = {
                'Rx': 'u3({}, -pi/2, pi/2)',
                'Ry': 'u3({}, 0, 0)',
                'Rz': 'u1({})'
            }
            u_name = {'Rx': 'u3', 'Ry': 'u3', 'Rz': 'u1'}
            u_angle = {
                'Rx': [gate.angle, -math.pi / 2, math.pi / 2],
                'Ry': [gate.angle, 0, 0],
                'Rz': [gate.angle]
            }
            gate_qasm = u_strs[str(gate)[0:2]].format(gate.angle)
            gate_name = u_name[str(gate)[0:2]]
            params = u_angle[str(gate)[0:2]]
            self.qasm += "\n{} q[{}];".format(gate_qasm, qb_pos)
            self._json.append({
                'qubits': [qb_pos],
                'name': gate_name,
                'params': params
            })
        elif gate == H:
            assert get_control_count(cmd) == 0
            qb_pos = cmd.qubits[0][0].id
            self.qasm += "\nu2(0,pi/2) q[{}];".format(qb_pos)
            self._json.append({
                'qubits': [qb_pos],
                'name': 'u2',
                'params': [0, 3.141592653589793]
            })
        else:
            raise Exception(
                'Command not authorized. You should run the circuit with the '
                'appropriate ibm setup.'
            )

    def _logical_to_physical(self, qb_id):
        """
        Return the physical location of the qubit with the given logical id.

        Args:
            qb_id (int): ID of the logical qubit whose position should be
                returned.
        """
        assert self.main_engine.mapper is not None
        mapping = self.main_engine.mapper.current_mapping
        if qb_id not in mapping:
            raise RuntimeError(
                "Unknown qubit id {}. Please make sure "
                "eng.flush() was called and that the qubit "
                "was eliminated during optimization.".format(qb_id))
        return mapping[qb_id]

    def get_probabilities(self, qureg):
        """
        Return the list of basis states with corresponding probabilities.
        If input qureg is a subset of the register used for the experiment,
        then returns the projected probabilities over the other states.

        The measured bits are ordered according to the supplied quantum
        register, i.e., the left-most bit in the state-string corresponds to
        the first qubit in the supplied quantum register.

        Warning:
            Only call this function after the circuit has been executed!

        Args:
            qureg (list<Qubit>): Quantum register determining the order of the
                qubits.

        Returns:
            probability_dict (dict): Dictionary mapping n-bit strings to
                probabilities.

        Raises:
            RuntimeError: If no data is available (i.e., if the circuit has
                not been executed). Or if a qubit was supplied which was not
                present in the circuit (might have gotten optimized away).
        """
        if len(self._probabilities) == 0:
            raise RuntimeError("Please, run the circuit first!")

        probability_dict = dict()
        for state in self._probabilities:
            mapped_state = ['0'] * len(qureg)
            for i in range(len(qureg)):
                mapped_state[i] = state[self._logical_to_physical(qureg[i].id)]
            probability = self._probabilities[state]
            mapped_state = "".join(mapped_state)
            if mapped_state not in probability_dict:
                probability_dict[mapped_state] = probability
            else:
                probability_dict[mapped_state] += probability
        return probability_dict

    def _run(self):
        """
        Run the circuit.

        Send the circuit via a non documented IBM API (using JSON written
        circuits) using the provided user data / ask for the user token.
        """
        # finally: add measurements (no intermediate measurements are allowed)
        for measured_id in self._measured_ids:
            qb_loc = self.main_engine.mapper.current_mapping[measured_id]
            self.qasm += "\nmeasure q[{}] -> c[{}];".format(qb_loc, qb_loc)
            self._json.append({
                'qubits': [qb_loc],
                'name': 'measure',
                'memory': [qb_loc]
            })
        # return if no operations / measurements have been performed.
        if self.qasm == "":
            return
        max_qubit_id = max(self._allocated_qubits) + 1
        info = {}
        info['json'] = self._json
        info['nq'] = max_qubit_id

        info['shots'] = self._num_runs
        info['maxCredits'] = 10
        info['backend'] = {'name': self.device}
        try:
            if self._retrieve_execution is None:
                res = send(info,
                           device=self.device,
                           token=self._token,
                           num_retries=self._num_retries,
                           interval=self._interval,
                           verbose=self._verbose)
            else:
                res = retrieve(device=self.device,
                               token=self._token,
                               jobid=self._retrieve_execution,
                               num_retries=self._num_retries,
                               interval=self._interval,
                               verbose=self._verbose)
            counts = res['data']['counts']
            # Determine random outcome
            P = random.random()
            p_sum = 0.
            measured = ""
            for state in counts:
                probability = counts[state] * 1. / self._num_runs
                state = "{0:b}".format(int(state, 0))
                state = state.zfill(max_qubit_id)
                # States in ibmq are right-ordered, so need to reverse state
                # string
                state = state[::-1]
                p_sum += probability
                star = ""
                if p_sum >= P and measured == "":
                    measured = state
                    star = "*"
                self._probabilities[state] = probability
                if self._verbose and probability > 0:
                    print(str(state) + " with p = " + str(probability) + star)

            class QB():
                def __init__(self, ID):
                    self.id = ID

            # register measurement result
            for ID in self._measured_ids:
                location = self._logical_to_physical(ID)
                result = int(measured[location])
                self.main_engine.set_measurement_result(QB(ID), result)
            self._reset()
        except TypeError:
            raise Exception("Failed to run the circuit. Aborting.")

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until
        completion.

        Args:
            command_list: List of commands to execute
        """
        for cmd in command_list:
            if not cmd.gate == FlushGate():
                self._store(cmd)
            else:
                self._run()
                self._reset()



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Defines the parent class from which all mappers should be derived.

There is only one engine currently allowed to be derived from
BasicMapperEngine. This allows the simulator to automatically translate
logical qubit ids to mapped ids.
"""
from copy import deepcopy

from . import BasicEngine, CommandModifier
from projectq.meta import drop_engine_after, insert_engine, LogicalQubitIDTag
from projectq.ops import MeasureGate


class BasicMapperEngine(BasicEngine):
    """
    Parent class for all Mappers.

    Attributes:
        self.current_mapping (dict): Keys are the logical qubit ids and values
                                     are the mapped qubit ids.

    """

    def __init__(self):
        BasicEngine.__init__(self)
        self._current_mapping = None

    @property
    def current_mapping(self):
        return deepcopy(self._current_mapping)

    @current_mapping.setter
    def current_mapping(self, current_mapping):
        self._current_mapping = current_mapping

    def _send_cmd_with_mapped_ids(self, cmd):
        """
        Send this Command using the mapped qubit ids of self.current_mapping.

        If it is a Measurement gate, then it adds a LogicalQubitID tag.

        Args:
            cmd: Command object with logical qubit ids.
        """
        new_cmd = deepcopy(cmd)
        qubits = new_cmd.qubits
        for qureg in qubits:
            for qubit in qureg:
                if qubit.id != -1:
                    qubit.id = self.current_mapping[qubit.id]
        control_qubits = new_cmd.control_qubits
        for qubit in control_qubits:
            qubit.id = self.current_mapping[qubit.id]
        if isinstance(new_cmd.gate, MeasureGate):
            assert len(new_cmd.qubits) == 1 and len(new_cmd.qubits[0]) == 1

            # Add LogicalQubitIDTag to MeasureGate
            def add_logical_id(command, old_tags=deepcopy(cmd.tags)):
                command.tags = (old_tags +
                                [LogicalQubitIDTag(cmd.qubits[0][0].id)])
                return command

            tagger_eng = CommandModifier(add_logical_id)
            insert_engine(self, tagger_eng)
            self.send([new_cmd])
            drop_engine_after(self)
        else:
            self.send([new_cmd])

    def receive(self, command_list):
        for cmd in command_list:
            self._send_cmd_with_mapped_ids(cmd)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   Module uses ideas from "Basic circuit compilation techniques for an
#   ion-trap quantum machine" by Dmitri Maslov (2017) at
#   https://iopscience.iop.org/article/10.1088/1367-2630/aa5e47
"""
Registers a decomposition for the Rz gate into an Rx and Ry(pi/2) or Ry(-pi/2)
gate
"""

import math

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Control, get_control_count, Uncompute
from projectq.ops import Rx, Ry, Rz


def _decompose_rz2rx_P(cmd):
    """ Decompose the Rz using negative angle. """
    # Labelled 'P' for 'plus' because decomposition ends with a Ry(+pi/2)
    qubit = cmd.qubits[0]
    eng = cmd.engine
    angle = cmd.gate.angle

    with Control(eng, cmd.control_qubits):
        with Compute(eng):
            Ry(-math.pi / 2.) | qubit
        Rx(-angle) | qubit
        Uncompute(eng)


def _decompose_rz2rx_M(cmd):
    """ Decompose the Rz using positive angle. """
    # Labelled 'M' for 'minus' because decomposition ends with a Ry(-pi/2)
    qubit = cmd.qubits[0]
    eng = cmd.engine
    angle = cmd.gate.angle

    with Control(eng, cmd.control_qubits):
        with Compute(eng):
            Ry(math.pi / 2.) | qubit
        Rx(angle) | qubit
        Uncompute(eng)


def _recognize_RzNoCtrl(cmd):
    """ Decompose the gate only if the command represents a single qubit gate
        (if it is not part of a control gate).
    """
    return get_control_count(cmd) == 0


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Rz, _decompose_rz2rx_P, _recognize_RzNoCtrl),
    DecompositionRule(Rz, _decompose_rz2rx_M, _recognize_RzNoCtrl)
]



import sys
from ._basics import BasicEngine, LastEngineException, ForwarderEngine

# Avoid dependency loop
setattr(sys.modules['projectq.cengines'], 'BasicEngine', BasicEngine)
setattr(sys.modules['projectq.cengines'], 'LastEngineException',
        LastEngineException)

from ._cmdmodifier import CommandModifier
from ._basicmapper import BasicMapperEngine
from ._swap_utils import return_swap_depth
from ._main import MainEngine, NotYetMeasuredError



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition for the controlled global phase gate.

Turns the controlled global phase gate into a (controlled) phase-shift gate.
Each time this rule is applied, one control can be shaved off.
"""

from projectq.cengines import DecompositionRule
from projectq.meta import Control, get_control_count
from projectq.ops import Ph, R


def _decompose_Ph(cmd):
    """ Decompose the controlled phase gate (C^nPh(phase)). """
    ctrl = cmd.control_qubits
    gate = cmd.gate
    eng = cmd.engine

    with Control(eng, ctrl[1:]):
        R(gate.angle) | ctrl[0]


def _recognize_Ph(cmd):
    """ Recognize the controlled phase gate. """
    return get_control_count(cmd) >= 1


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Ph, _decompose_Ph, _recognize_Ph)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a CommandModifier engine, which can be used to, e.g., modify the tags
of all commands which pass by (see the AutoReplacer for an example).
"""
from ._basics import BasicEngine


class CommandModifier(BasicEngine):
    """
    CommandModifier is a compiler engine which applies a function to all
    incoming commands, sending on the resulting command instead of the
    original one.
    """
    def __init__(self, cmd_mod_fun):
        """
        Initialize the CommandModifier.

        Args:
            cmd_mod_fun (function): Function which, given a command cmd,
                returns the command it should send instead.

        Example:
            .. code-block:: python

                def cmd_mod_fun(cmd):
                    cmd.tags += [MyOwnTag()]
                compiler_engine = CommandModifier(cmd_mod_fun)
                ...
        """
        BasicEngine.__init__(self)
        self._cmd_mod_fun = cmd_mod_fun

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, modify all
        commands, and send them on to the next engine.

        Args:
            command_list (list<Command>): List of commands to receive and then
                (after modification) send on.
        """
        new_command_list = [self._cmd_mod_fun(cmd) for cmd in command_list]
        self.send(new_command_list)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Contains a compiler engine to add mapping information
"""
from projectq.cengines import BasicMapperEngine


class ManualMapper(BasicMapperEngine):
    """
    Manual Mapper which adds QubitPlacementTags to Allocate gate commands
    according to a user-specified mapping.

    Attributes:
        map (function): The function which maps a given qubit id to its
            location. It gets set when initializing the mapper.
    """

    def __init__(self, map_fun=lambda x: x):
        """
        Initialize the mapper to a given mapping. If no mapping function is
        provided, the qubit id is used as the location.

        Args:
            map_fun (function): Function which, given the qubit id, returns
                an integer describing the physical location (must be constant).
        """
        BasicMapperEngine.__init__(self)
        self.map = map_fun
        self.current_mapping = dict()

    def receive(self, command_list):
        """
        Receives a command list and passes it to the next engine, adding
        qubit placement tags to allocate gates.

        Args:
            command_list (list of Command objects): list of commands to
                receive.
        """
        for cmd in command_list:
            ids = [qb.id for qr in cmd.qubits for qb in qr]
            ids += [qb.id for qb in cmd.control_qubits]
            for ID in ids:
                if ID not in self.current_mapping:
                    self._current_mapping[ID] = self.map(ID)
            self._send_cmd_with_mapped_ids(cmd)



# Some of the setup.py code is inspired or copied from SQLAlchemy

# SQLAlchemy was created by Michael Bayer.

# Major contributing authors include:

# - Michael Bayer <mike_mp@zzzcomputing.com>
# - Jason Kirtland <jek@discorporate.us>
# - Gaetan de Menten <gdementen@gmail.com>
# - Diana Clarke <diana.joan.clarke@gmail.com>
# - Michael Trier <mtrier@gmail.com>
# - Philip Jenvey <pjenvey@underboss.org>
# - Ants Aasma <ants.aasma@gmail.com>
# - Paul Johnston <paj@pajhome.org.uk>
# - Jonathan Ellis <jbellis@gmail.com>
# - Damien Nguyen <damien1@huawei.com> (ProjectQ)

# Copyright 2005-2020 SQLAlchemy and ProjectQ authors and contributors (see above)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from __future__ import print_function
from distutils.errors import (CompileError, LinkError, CCompilerError,
                              DistutilsExecError, DistutilsPlatformError)
import setuptools
from setuptools.command.build_ext import build_ext
from setuptools.command.egg_info import egg_info
import sys
import os
import subprocess
import platform

# ==============================================================================
# Helper functions and classes


class get_pybind_include(object):
    '''Helper class to determine the pybind11 include path

    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. '''
    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def important_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


def status_msgs(*msgs):
    print('-' * 75)
    for msg in msgs:
        print('# INFO: ', msg)
    print('-' * 75)


def compiler_test(compiler,
                  flagname=None,
                  link=False,
                  include='',
                  body='',
                  postargs=None):
    '''
    Return a boolean indicating whether a flag name is supported on the
    specified compiler.
    '''
    import tempfile
    f = tempfile.NamedTemporaryFile('w', suffix='.cpp', delete=False)
    f.write('{}\nint main (int argc, char **argv) {{ {} return 0; }}'.format(
        include, body))
    f.close()
    ret = True

    if postargs is None:
        postargs = [flagname] if flagname is not None else None
    elif flagname is not None:
        postargs.append(flagname)

    try:
        exec_name = os.path.join(tempfile.mkdtemp(), 'test')

        if compiler.compiler_type == 'msvc':
            olderr = os.dup(sys.stderr.fileno())
            err = open('err.txt', 'w')
            os.dup2(err.fileno(), sys.stderr.fileno())

        obj_file = compiler.compile([f.name], extra_postargs=postargs)
        if not os.path.exists(obj_file[0]):
            raise RuntimeError('')
        if link:
            compiler.link_executable(obj_file,
                                     exec_name,
                                     extra_postargs=postargs)

        if compiler.compiler_type == 'msvc':
            err.close()
            os.dup2(olderr, sys.stderr.fileno())
            with open('err.txt', 'r') as err_file:
                if err_file.readlines():
                    raise RuntimeError('')
    except (CompileError, LinkError, RuntimeError):
        ret = False
    os.unlink(f.name)
    return ret


def _fix_macosx_header_paths(*args):
    # Fix path to SDK headers if necessary
    _MACOSX_XCODE_REF_PATH = ('/Applications/Xcode.app/Contents/'
                              + 'Developer/Platforms/MacOSX.platform/'
                              + 'Developer')
    _MACOSX_DEVTOOLS_REF_PATH = '/Library/Developer/CommandLineTools/'
    _has_xcode = os.path.exists(_MACOSX_XCODE_REF_PATH)
    _has_devtools = os.path.exists(_MACOSX_DEVTOOLS_REF_PATH)
    if not _has_xcode and not _has_devtools:
        important_msgs('ERROR: Must install either Xcode or '
                       + 'CommandLineTools!')
        raise BuildFailed()

    def _do_replace(idx, item):
        if not _has_xcode and _MACOSX_XCODE_REF_PATH in item:
            compiler_args[idx] = item.replace(_MACOSX_XCODE_REF_PATH,
                                              _MACOSX_DEVTOOLS_REF_PATH)

        if not _has_devtools and _MACOSX_DEVTOOLS_REF_PATH in item:
            compiler_args[idx] = item.replace(_MACOSX_DEVTOOLS_REF_PATH,
                                              _MACOSX_XCODE_REF_PATH)

    for compiler_args in args:
        for idx, item in enumerate(compiler_args):
            _do_replace(idx, item)


# ------------------------------------------------------------------------------


class BuildFailed(Exception):
    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax


# ------------------------------------------------------------------------------
# Python build related variable

cpython = platform.python_implementation() == 'CPython'
ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError, )

# ==============================================================================

# This package's name
pkg_name = 'hiq-projectq'

# This reads the __version__ variable from projectq/_version.py
exec(open('projectq/_version.py').read())

# Readme file as long_description:
long_description = open('README.rst').read()

# Read in requirements.txt
with open('requirements.txt', 'r') as f_requirements:
    requirements = f_requirements.readlines()
requirements = [r.strip() for r in requirements]

# ------------------------------------------------------------------------------
# ProjectQ C++ extensions

ext_modules = [
    setuptools.Extension(
        'projectq.backends._sim._cppsim',
        ['projectq/backends/_sim/_cppsim.cpp'],
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True)
        ],
        language='c++'),
]

# ==============================================================================


class BuildExt(build_ext):
    '''A custom build extension for adding compiler-specific options.'''
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            raise BuildFailed()

    def build_extensions(self):
        self._configure_compiler()
        for ext in self.extensions:
            ext.extra_compile_args = self.opts
            ext.extra_link_args = self.link_opts
        try:
            build_ext.build_extensions(self)
        except ext_errors:
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise

    def _configure_compiler(self):
        if sys.platform == 'darwin':
            _fix_macosx_header_paths(self.compiler.compiler,
                                     self.compiler.compiler_so)

            if compiler_test(self.compiler, '-stdlib=libc++'):
                self.c_opts['unix'] += ['-stdlib=libc++']

        ct = self.compiler.compiler_type
        self.opts = self.c_opts.get(ct, [])
        self.link_opts = []

        if not compiler_test(self.compiler):
            important_msgs(
                'ERROR: something is wrong with your C++ compiler.\n'
                'Failed to compile a simple test program!')
            raise BuildFailed()

        # ------------------------------

        status_msgs('Configuring OpenMP')
        self._configure_openmp()
        status_msgs('Configuring compiler intrinsics')
        self._configure_intrinsics()
        status_msgs('Configuring C++ standard')
        self._configure_cxx_standard()

        # ------------------------------
        # Other compiler tests

        status_msgs('Other compiler tests')
        if ct == 'unix':
            if compiler_test(self.compiler, '-fvisibility=hidden'):
                self.opts.append('-fvisibility=hidden')
            self.opts.append("-DVERSION_INFO=\"{}\"".format(
                self.distribution.get_version()))
        elif ct == 'msvc':
            self.opts.append("/DVERSION_INFO=\\'{}\\'".format(
                self.distribution.get_version()))

        status_msgs('Finished configuring compiler!')

    def _configure_openmp(self):
        if self.compiler.compiler_type == 'msvc':
            return

        kwargs = {
            'link': True,
            'include': '#include <omp.h>',
            'body': 'int a = omp_get_num_threads(); ++a;'
        }

        for flag in ['-openmp', '-fopenmp', '-qopenmp', '/Qopenmp']:
            if compiler_test(self.compiler, flag, **kwargs):
                self.opts.append(flag)
                self.link_opts.append(flag)
                return

        flag = '-fopenmp'
        if (sys.platform == 'darwin' and compiler_test(self.compiler, flag)):
            try:
                llvm_root = subprocess.check_output(
                    ['brew', '--prefix', 'llvm']).decode('utf-8')[:-1]
                compiler_root = subprocess.check_output(
                    ['which', self.compiler.compiler[0]]).decode('utf-8')[:-1]

                # Only add the flag if the compiler we are using is the one
                # from HomeBrew
                if llvm_root in compiler_root:
                    l_arg = '-L{}/lib'.format(llvm_root)
                    if compiler_test(self.compiler,
                                     flag,
                                     postargs=[l_arg],
                                     **kwargs):
                        self.opts.append(flag)
                        self.link_opts.extend((l_arg, flag))
                        return
            except subprocess.CalledProcessError:
                pass

            try:
                # Only relevant for MacPorts users with clang-3.7
                port_path = subprocess.check_output(['which', 'port'
                                                     ]).decode('utf-8')[:-1]
                macports_root = os.path.dirname(os.path.dirname(port_path))
                compiler_root = subprocess.check_output(
                    ['which', self.compiler.compiler[0]]).decode('utf-8')[:-1]

                # Only add the flag if the compiler we are using is the one
                # from MacPorts
                if macports_root in compiler_root:
                    c_arg = '-I{}/include/libomp'.format(macports_root)
                    l_arg = '-L{}/lib/libomp'.format(macports_root)

                    if compiler_test(self.compiler,
                                     flag,
                                     postargs=[c_arg, l_arg],
                                     **kwargs):
                        self.opts.extend((c_arg, flag))
                        self.link_opts.extend((l_arg, flag))
                        return
            except subprocess.CalledProcessError:
                pass

        important_msgs('WARNING: compiler does not support OpenMP!')

    def _configure_intrinsics(self):
        for flag in [
                '-march=native', '-mavx2', '/arch:AVX2', '/arch:CORE-AVX2',
                '/arch:AVX'
        ]:
            if compiler_test(
                    self.compiler,
                    flagname=flag,
                    link=False,
                    include='#include <immintrin.h>',
                    body='__m256d neg = _mm256_set1_pd(1.0); (void)neg;'):

                if sys.platform == 'win32':
                    self.opts.extend(('/DINTRIN', flag))
                else:
                    self.opts.extend(('-DINTRIN', flag))
                break

        for flag in ['-ffast-math', '-fast', '/fast', '/fp:precise']:
            if compiler_test(self.compiler, flagname=flag):
                self.opts.append(flag)
                break

    def _configure_cxx_standard(self):
        if self.compiler.compiler_type == 'msvc':
            return

        cxx_standards = [17, 14, 11]
        if sys.version_info[0] < 3:
            cxx_standards = [year for year in cxx_standards if year < 17]

        if sys.platform == 'darwin':
            _, minor_version, _ = [
                int(i) for i in platform.mac_ver()[0].split('.')
            ]
            if minor_version < 14:
                cxx_standards = [year for year in cxx_standards if year < 17]

        for year in cxx_standards:
            flag = '-std=c++{}'.format(year)
            if compiler_test(self.compiler, flag):
                self.opts.append(flag)
                return
            flag = '/Qstd=c++{}'.format(year)
            if compiler_test(self.compiler, flag):
                self.opts.append(flag)
                return

        important_msgs('ERROR: compiler needs to have at least C++11 support!')
        raise BuildFailed()


class EggInfo(egg_info):
    def run(self):
        try:
            with open(os.devnull, 'w') as devnull:
                subprocess.check_call(
                    ['python3', '-m', 'pip', 'show', 'projectq'],
                    stdout=devnull,
                    stderr=devnull)

            raise RuntimeError(
                ('Error: unable to install {0} {1} because you have ProjectQ '
                 'installed.\n'
                 'Please uninstall ProjectQ before trying to install {0} '
                 'again.\nNote that {0} has the same functionality as '
                 'ProjectQ and you will be able to run `import projectq` '
                 'statements as usual').format(pkg_name, __version__))
        except subprocess.CalledProcessError:
            pass
        egg_info.run(self)


class Distribution(setuptools.Distribution):
    def has_ext_modules(self):
        # We want to always claim that we have ext_modules. This will be fine
        # if we don't actually have them (such as on PyPy) because nothing
        # will get built, however we don't want to provide an overally broad
        # Wheel package when building a wheel without C support. This will
        # ensure that Wheel knows to treat us as if the build output is
        # platform specific.
        return True


# ==============================================================================


def run_setup(with_cext):
    kwargs = {}
    if with_cext:
        kwargs['ext_modules'] = ext_modules
    else:
        kwargs['ext_modules'] = []

    setuptools.setup(
        name=pkg_name,
        version=__version__,
        author='ProjectQ',
        author_email='info@projectq.ch',
        url='http://www.projectq.ch',
        project_urls={
            'Documentation': 'https://projectq.readthedocs.io/en/latest/',
            'Issue Tracker':
            'https://github.com/ProjectQ-Framework/ProjectQ/',
        },
        description=('ProjectQ - An open source software framework for '
                     'quantum computing'),
        long_description=long_description,
        install_requires=requirements,
        cmdclass={'build_ext': BuildExt, 'egg_info': EggInfo},
        zip_safe=False,
        license='Apache 2',
        packages=setuptools.find_packages(),
        distclass=Distribution,
        **kwargs)


# ==============================================================================

if not cpython:
    run_setup(False)
    important_msgs(
        'WARNING: C/C++ extensions are not supported on '
        + 'some features are disabled (e.g. C++ simulator).',
        'Plain-Python build succeeded.',
    )
elif os.environ.get('DISABLE_PROJECTQ_CEXT'):
    run_setup(False)
    important_msgs(
        'DISABLE_PROJECTQ_CEXT is set; '
        + 'not attempting to build C/C++ extensions.',
        'Plain-Python build succeeded.',
    )

else:
    try:
        run_setup(True)
    except BuildFailed as exc:
        important_msgs(
            exc.cause,
            'WARNING: Some C/C++ extensions could not be compiled, '
            + 'some features are disabled (e.g. C++ simulator).',
            'Failure information, if any, is above.',
            'Retrying the build without the C/C++ extensions now.',
        )

        run_setup(False)

        important_msgs(
            'WARNING: Some C/C++ extensions could not be compiled, '
            + 'some features are disabled (e.g. C++ simulator).',
            'Plain-Python build succeeded.',
        )



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
This file defines the apply_command function and the Command class.

When a gate is applied to qubits, e.g.,

.. code-block:: python

    CNOT | (qubit1, qubit2)

a Command object is generated which represents both the gate, qubits and
control qubits. This Command object then gets sent down the compilation
pipeline.

In detail, the Gate object overloads the operator| (magic method __or__)
to generate a Command object which stores the qubits in a canonical order
using interchangeable qubit indices defined by the gate to allow the
optimizer to cancel the following two gates

.. code-block:: python

    Swap | (qubit1, qubit2)
    Swap | (qubit2, qubit1)

The command then gets sent to the MainEngine via the
apply wrapper (apply_command).
"""

from copy import deepcopy

import projectq
from projectq.types import WeakQubitRef, Qureg


def apply_command(cmd):
    """
    Apply a command.

    Extracts the qubits-owning (target) engine from the Command object
    and sends the Command to it.

    Args:
        cmd (Command): Command to apply
    """
    engine = cmd.engine
    engine.receive([cmd])


class Command(object):
    """
    Class used as a container to store commands. If a gate is applied to
    qubits, then the gate and qubits are saved in a command object. Qubits
    are copied into WeakQubitRefs in order to allow early deallocation (would
    be kept alive otherwise). WeakQubitRef qubits don't send deallocate gate
    when destructed.

    Attributes:
        gate: The gate to execute
        qubits: Tuple of qubit lists (e.g. Quregs). Interchangeable qubits
                  are stored in a unique order
        control_qubits: The Qureg of control qubits in a unique order
        engine: The engine (usually: MainEngine)
        tags: The list of tag objects associated with this command
          (e.g., ComputeTag, UncomputeTag, LoopTag, ...). tag objects need to
          support ==, != (__eq__ and __ne__) for comparison as used in e.g.
          TagRemover. New tags should always be added to the end of the list.
          This means that if there are e.g. two LoopTags in a command, tag[0]
          is from the inner scope while tag[1] is from the other scope as the
          other scope receives the command after the inner scope LoopEngine
          and hence adds its LoopTag to the end.
        all_qubits: A tuple of control_qubits + qubits
    """
    def __init__(self, engine, gate, qubits, controls=(), tags=()):
        """
        Initialize a Command object.

        Note:
            control qubits (Command.control_qubits) are stored as a
            list of qubits, and command tags (Command.tags) as a list of tag-
            objects. All functions within this class also work if
            WeakQubitRefs are supplied instead of normal Qubit objects
            (see WeakQubitRef).

        Args:
            engine (projectq.cengines.BasicEngine):
                engine which created the qubit (mostly the MainEngine)
            gate (projectq.ops.Gate):
                Gate to be executed
            qubits (tuple[Qureg]):
                Tuple of quantum registers (to which the gate is applied)
            controls (Qureg|list[Qubit]):
                Qubits that condition the command.
            tags (list[object]):
                Tags associated with the command.
        """

        qubits = tuple(
            [WeakQubitRef(qubit.engine, qubit.id) for qubit in qreg]
            for qreg in qubits)

        self.gate = gate
        self.tags = list(tags)
        self.qubits = qubits  # property
        self.control_qubits = controls  # property
        self.engine = engine  # property

    @property
    def qubits(self):
        return self._qubits

    @qubits.setter
    def qubits(self, qubits):
        self._qubits = self._order_qubits(qubits)

    def __deepcopy__(self, memo):
        """ Deepcopy implementation. Engine should stay a reference."""
        return Command(self.engine, deepcopy(self.gate), self.qubits,
                       list(self.control_qubits), deepcopy(self.tags))

    def get_inverse(self):
        """
        Get the command object corresponding to the inverse of this command.

        Inverts the gate (if possible) and creates a new command object from
        the result.

        Raises:
            NotInvertible: If the gate does not provide an inverse (see
                BasicGate.get_inverse)
        """
        return Command(self._engine, projectq.ops.get_inverse(self.gate),
                       self.qubits, list(self.control_qubits),
                       deepcopy(self.tags))

    def is_identity(self):
        """
        Evaluate if the gate called in the command object is an identity gate.

        Returns:
            True if the gate is equivalent to an Identity gate, False otherwise
        """
        return projectq.ops.is_identity(self.gate)

    def get_merged(self, other):
        """
        Merge this command with another one and return the merged command
        object.

        Args:
            other: Other command to merge with this one (self)

        Raises:
            NotMergeable: if the gates don't supply a get_merged()-function
                or can't be merged for other reasons.
        """
        if (self.tags == other.tags and self.all_qubits == other.all_qubits
                and self.engine == other.engine):
            return Command(self.engine, self.gate.get_merged(other.gate),
                           self.qubits, self.control_qubits,
                           deepcopy(self.tags))
        raise projectq.ops.NotMergeable("Commands not mergeable.")

    def _order_qubits(self, qubits):
        """
        Order the given qubits according to their IDs (for unique comparison of
        commands).

        Args:
            qubits: Tuple of quantum registers (i.e., tuple of lists of qubits)

        Returns: Ordered tuple of quantum registers
        """
        ordered_qubits = list(qubits)
        # e.g. [[0,4],[1,2,3]]
        interchangeable_qubit_indices = self.interchangeable_qubit_indices
        for old_positions in interchangeable_qubit_indices:
            new_positions = sorted(old_positions,
                                   key=lambda x: ordered_qubits[x][0].id)
            qubits_new_order = [ordered_qubits[i] for i in new_positions]
            for i in range(len(old_positions)):
                ordered_qubits[old_positions[i]] = qubits_new_order[i]
        return tuple(ordered_qubits)

    @property
    def interchangeable_qubit_indices(self):
        """
        Return nested list of qubit indices which are interchangeable.

        Certain qubits can be interchanged (e.g., the qubit order for a Swap
        gate). To ensure that only those are sorted when determining the
        ordering (see _order_qubits), self.interchangeable_qubit_indices is
        used.
        Example:
            If we can interchange qubits 0,1 and qubits 3,4,5,
            then this function returns [[0,1],[3,4,5]]
        """
        return self.gate.interchangeable_qubit_indices

    @property
    def control_qubits(self):
        """ Returns Qureg of control qubits."""
        return self._control_qubits

    @control_qubits.setter
    def control_qubits(self, qubits):
        """
        Set control_qubits to qubits

        Args:
            control_qubits (Qureg): quantum register
        """
        self._control_qubits = ([
            WeakQubitRef(qubit.engine, qubit.id) for qubit in qubits
        ])
        self._control_qubits = sorted(self._control_qubits, key=lambda x: x.id)

    def add_control_qubits(self, qubits):
        """
        Add (additional) control qubits to this command object.

        They are sorted to ensure a canonical order. Also Qubit objects
        are converted to WeakQubitRef objects to allow garbage collection and
        thus early deallocation of qubits.

        Args:
            qubits (list of Qubit objects): List of qubits which control this
                gate, i.e., the gate is only executed if all qubits are
                in state 1.
        """
        assert (isinstance(qubits, list))
        self._control_qubits.extend(
            [WeakQubitRef(qubit.engine, qubit.id) for qubit in qubits])
        self._control_qubits = sorted(self._control_qubits, key=lambda x: x.id)

    @property
    def all_qubits(self):
        """
        Get all qubits (gate and control qubits).

        Returns a tuple T where T[0] is a quantum register (a list of
        WeakQubitRef objects) containing the control qubits and T[1:] contains
        the quantum registers to which the gate is applied.
        """
        return (self.control_qubits, ) + self.qubits

    @property
    def engine(self):
        """
        Return engine to which the qubits belong / on which the gates are
        executed.
        """
        return self._engine

    @engine.setter
    def engine(self, engine):
        """
        Set / Change engine of all qubits to engine.

        Args:
            engine: New owner of qubits and owner of this Command object
        """
        self._engine = engine
        for qureg in self.qubits:
            for qubit in qureg:
                qubit.engine = engine
        for qubit in self.control_qubits:
            qubit.engine = engine

    def __eq__(self, other):
        """
        Compare this command to another command.

        Args:
            other (Command): Command object to compare this to

        Returns: True if Command objects are equal (same gate, applied to same
        qubits; ordered modulo interchangeability; and same tags)
        """
        if (isinstance(other, self.__class__) and self.gate == other.gate
                and self.tags == other.tags and self.engine == other.engine
                and self.all_qubits == other.all_qubits):
            return True
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.to_string()

    def to_string(self, symbols=False):
        """
        Get string representation of this Command object.
        """
        qubits = self.qubits
        ctrlqubits = self.control_qubits
        if len(ctrlqubits) > 0:
            qubits = (self.control_qubits, ) + qubits
        qstring = ""
        if len(qubits) == 1:
            qstring = str(Qureg(qubits[0]))
        else:
            qstring = "( "
            for qreg in qubits:
                qstring += str(Qureg(qreg))
                qstring += ", "
            qstring = qstring[:-2] + " )"
        cstring = "C" * len(ctrlqubits)
        return cstring + self.gate.to_string(symbols) + " | " + qstring


all_defined_symbols = [apply_command, Command]



#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import matplotlib.pyplot as plt


def histogram(sim, qureg):
    """
    Make a measurement outcome probability histogram for the given qubits.

    Args:
        sim (Simulator): The simulator to call get_probability
        qureg (list of qubits and/or quregs): The qubits,
            for which to make the histogram

    Returns:
        A tuple (fig, axes, probabilities), where:
        fig: The histogram as figure
        axes: The axes of the histogram
        probabilities (dict): A dictionary mapping outcomes as string
        to their probabilities

    Note:
        Don't forget to call eng.flush() before using this function.
    """
    qubit_list = []
    for q in qureg:
        if isinstance(q, list):
            qubit_list.extend(q)
        else:
            qubit_list.append(q)

    if len(qubit_list) > 5:
        print('Warning: For {0} qubits there are 2^{0} different outcomes'.
              format(len(qubit_list)))
        print("The resulting histogram may look bad and/or take too long.")
        print("Consider calling histogram() with a sublist of the qubits.")

    outcome = [0] * len(qubit_list)
    n_outcomes = (1 << len(qubit_list))
    probabilities = {}
    for i in range(n_outcomes):
        for pos in range(len(qubit_list)):
            if (1 << pos) & i:
                outcome[pos] = 1
            else:
                outcome[pos] = 0
        probabilities[''.join([str(bit) for bit in outcome
                               ])] = sim.get_probability(outcome, qubit_list)

    # Empirical figure size for up to 5 qubits
    fig, axes = plt.subplots(figsize=(min(21.2, 2
                                          + 0.6 * (1 << len(qubit_list))), 7))
    names = list(probabilities.keys())
    values = list(probabilities.values())
    axes.bar(names, values)
    fig.suptitle('Measurement Probabilities')
    return (fig, axes, probabilities)



#   Copyright 2018 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from ._basics import BasicGate


class StatePreparation(BasicGate):
    """
    Gate for transforming qubits in state |0> to any desired quantum state.
    """
    def __init__(self, final_state):
        """
        Initialize StatePreparation gate.

        Example:
            .. code-block:: python

                qureg = eng.allocate_qureg(2)
                StatePreparation([0.5, -0.5j, -0.5, 0.5]) | qureg

        Note:
            final_state[k] is taken to be the amplitude of the computational
            basis state whose string is equal to the binary representation
            of k.

        Args:
            final_state(list[complex]): wavefunction of the desired
                                        quantum state. len(final_state) must
                                        be 2**len(qureg). Must be normalized!
        """
        BasicGate.__init__(self)
        self.final_state = list(final_state)

    def __str__(self):
        return "StatePreparation"

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.final_state == other.final_state
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash("StatePreparation(" + str(self.final_state) + ")")



#   Copyright 2020 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains a compiler engine which generates matplotlib figures describing the
circuit.
"""

from builtins import input
import re
import itertools

from projectq.cengines import LastEngineException, BasicEngine
from projectq.ops import (FlushGate, Measure, Allocate, Deallocate)
from projectq.meta import get_control_count
from projectq.backends._circuits import to_draw

# ==============================================================================


def _format_gate_str(cmd):
    param_str = ''
    gate_name = str(cmd.gate)
    if '(' in gate_name:
        (gate_name, param_str) = re.search(r'(.+)\((.*)\)', gate_name).groups()
        params = re.findall(r'([^,]+)', param_str)
        params_str_list = []
        for param in params:
            try:
                params_str_list.append('{0:.2f}'.format(float(param)))
            except ValueError:
                if len(param) < 8:
                    params_str_list.append(param)
                else:
                    params_str_list.append(param[:5] + '...')

        gate_name += '(' + ','.join(params_str_list) + ')'
    return gate_name


# ==============================================================================


class CircuitDrawerMatplotlib(BasicEngine):
    """
    CircuitDrawerMatplotlib is a compiler engine which using Matplotlib library
    for drawing quantum circuits
    """
    def __init__(self, accept_input=False, default_measure=0):
        """
        Initialize a circuit drawing engine(mpl)
        Args:
            accept_input (bool): If accept_input is true, the printer queries
                the user to input measurement results if the CircuitDrawerMPL
                is the last engine. Otherwise, all measurements yield the
                result default_measure (0 or 1).
            default_measure (bool): Default value to use as measurement
                results if accept_input is False and there is no underlying
                backend to register real measurement results.
        """
        BasicEngine.__init__(self)
        self._accept_input = accept_input
        self._default_measure = default_measure
        self._map = dict()
        self._qubit_lines = {}

    def is_available(self, cmd):
        """
        Specialized implementation of is_available: Returns True if the
        CircuitDrawerMatplotlib is the last engine
        (since it can print any command).

        Args:
            cmd (Command): Command for which to check availability (all
                Commands can be printed).

        Returns:
            availability (bool): True, unless the next engine cannot handle
            the Command (if there is a next engine).
        """
        try:
            # Multi-qubit gates may fail at drawing time if the target qubits
            # are not right next to each other on the output graphic.
            return BasicEngine.is_available(self, cmd)
        except LastEngineException:
            return True

    def _process(self, cmd):
        """
        Process the command cmd and stores it in the internal storage

        Queries the user for measurement input if a measurement command
        arrives if accept_input was set to True. Otherwise, it uses the
        default_measure parameter to register the measurement outcome.

        Args:
            cmd (Command): Command to add to the circuit diagram.
        """
        if cmd.gate == Allocate:
            qubit_id = cmd.qubits[0][0].id
            if qubit_id not in self._map:
                self._map[qubit_id] = qubit_id
            self._qubit_lines[qubit_id] = []
            return

        if cmd.gate == Deallocate:
            return

        if self.is_last_engine and cmd.gate == Measure:
            assert get_control_count(cmd) == 0
            for qureg in cmd.qubits:
                for qubit in qureg:
                    if self._accept_input:
                        measurement = None
                        while measurement not in ('0', '1', 1, 0):
                            prompt = ("Input measurement result (0 or 1) for "
                                      "qubit " + str(qubit) + ": ")
                            measurement = input(prompt)
                    else:
                        measurement = self._default_measure
                    self.main_engine.set_measurement_result(
                        qubit, int(measurement))

        targets = [qubit.id for qureg in cmd.qubits for qubit in qureg]
        controls = [qubit.id for qubit in cmd.control_qubits]

        ref_qubit_id = targets[0]
        gate_str = _format_gate_str(cmd)

        # First find out what is the maximum index that this command might
        # have
        max_depth = max(
            len(self._qubit_lines[qubit_id])
            for qubit_id in itertools.chain(targets, controls))

        # If we have a multi-qubit gate, make sure that all the qubit axes
        # have the same depth. We do that by recalculating the maximum index
        # over all the known qubit axes.
        # This is to avoid the possibility of a multi-qubit gate overlapping
        # with some other gates. This could potentially be improved by only
        # considering the qubit axes that are between the topmost and
        # bottommost qubit axes of the current command.
        if len(targets) + len(controls) > 1:
            max_depth = max(
                len(self._qubit_lines[qubit_id])
                for qubit_id in self._qubit_lines)

        for qubit_id in itertools.chain(targets, controls):
            depth = len(self._qubit_lines[qubit_id])
            self._qubit_lines[qubit_id] += [None] * (max_depth - depth)

            if qubit_id == ref_qubit_id:
                self._qubit_lines[qubit_id].append(
                    (gate_str, targets, controls))
            else:
                self._qubit_lines[qubit_id].append(None)

    def receive(self, command_list):
        """
        Receive a list of commands from the previous engine, print the
        commands, and then send them on to the next engine.

        Args:
            command_list (list<Command>): List of Commands to print (and
                potentially send on to the next engine).
        """
        for cmd in command_list:
            if not isinstance(cmd.gate, FlushGate):
                self._process(cmd)

            if not self.is_last_engine:
                self.send([cmd])

    def draw(self, qubit_labels=None, drawing_order=None, **kwargs):
        """
        Generates and returns the plot of the quantum circuit stored so far

        Args:
            qubit_labels (dict): label for each wire in the output figure.
                Keys: qubit IDs, Values: string to print out as label for
                that particular qubit wire.
            drawing_order (dict): position of each qubit in the output
                graphic. Keys: qubit IDs, Values: position of qubit on the
                qubit line in the graphic.
            **kwargs (dict): additional parameters are used to update
                the default plot parameters

        Returns:
            A tuple containing the matplotlib figure and axes objects

        Note:
            Additional keyword arguments can be passed to this
            function in order to further customize the figure output
            by matplotlib (default value in parentheses):

              - fontsize (14): Font size in pt
              - column_spacing (.5): Vertical spacing between two
                neighbouring gates (roughly in inches)
              - control_radius (.015): Radius of the circle for controls
              - labels_margin (1): Margin between labels and begin of
                wire (roughly in inches)
              - linewidth (1): Width of line
              - not_radius (.03): Radius of the circle for X/NOT gates
              - gate_offset (.05): Inner margins for gates with a text
                representation
              - mgate_width (.1): Width of the measurement gate
              - swap_delta (.02): Half-size of the SWAP gate
              - x_offset (.05): Absolute X-offset for drawing within the axes
              - wire_height (1): Vertical spacing between two qubit
                wires (roughly in inches)
        """
        max_depth = max(
            len(self._qubit_lines[qubit_id]) for qubit_id in self._qubit_lines)
        for qubit_id in self._qubit_lines:
            depth = len(self._qubit_lines[qubit_id])
            if depth < max_depth:
                self._qubit_lines[qubit_id] += [None] * (max_depth - depth)

        return to_draw(self._qubit_lines,
                       qubit_labels=qubit_labels,
                       drawing_order=drawing_order,
                       **kwargs)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Defines the BasicGate class, the base class of all gates, the
BasicRotationGate class, the SelfInverseGate, the FastForwardingGate, the
ClassicalInstruction gate, and the BasicMathGate class.

Gates overload the | operator to allow the following syntax:

.. code-block:: python

    Gate | (qureg1, qureg2, qureg2)
    Gate | (qureg, qubit)
    Gate | qureg
    Gate | qubit
    Gate | (qubit,)

This means that for more than one quantum argument (right side of | ), a tuple
needs to be made explicitely, while for one argument it is optional.
"""

import math
from copy import deepcopy

import numpy as np

from projectq.types import BasicQubit
from ._command import Command, apply_command

import unicodedata

ANGLE_PRECISION = 12
ANGLE_TOLERANCE = 10**-ANGLE_PRECISION
RTOL = 1e-10
ATOL = 1e-12


class NotMergeable(Exception):
    """
    Exception thrown when trying to merge two gates which are not mergeable (or
    where it is not    implemented (yet)).
    """
    pass


class NotInvertible(Exception):
    """
    Exception thrown when trying to invert a gate which is not invertable (or
    where the inverse is not  implemented (yet)).
    """
    pass


class BasicGate(object):
    """
    Base class of all gates. (Don't use it directly but derive from it)
    """
    def __init__(self):
        """
        Initialize a basic gate.

        Note:
            Set interchangeable qubit indices!
            (gate.interchangeable_qubit_indices)

            As an example, consider

            .. code-block:: python

                ExampleGate | (a,b,c,d,e)

            where a and b are interchangeable. Then, call this function as
            follows:

            .. code-block:: python

                self.set_interchangeable_qubit_indices([[0,1]])

            As another example, consider

            .. code-block:: python

                ExampleGate2 | (a,b,c,d,e)

            where a and b are interchangeable and, in addition, c, d, and e
            are interchangeable among themselves. Then, call this function as

            .. code-block:: python

                self.set_interchangeable_qubit_indices([[0,1],[2,3,4]])
        """
        self.interchangeable_qubit_indices = []

    def get_inverse(self):
        """
        Return the inverse gate.

        Standard implementation of get_inverse:

        Raises:
            NotInvertible: inverse is not implemented
        """
        raise NotInvertible("BasicGate: No get_inverse() implemented.")

    def get_merged(self, other):
        """
        Return this gate merged with another gate.

        Standard implementation of get_merged:

        Raises:
            NotMergeable: merging is not implemented
        """
        raise NotMergeable("BasicGate: No get_merged() implemented.")

    @staticmethod
    def make_tuple_of_qureg(qubits):
        """
        Convert quantum input of "gate | quantum input" to internal formatting.

        A Command object only accepts tuples of Quregs (list of Qubit objects)
        as qubits input parameter. However, with this function we allow the
        user to use a more flexible syntax:

            1) Gate | qubit
            2) Gate | [qubit0, qubit1]
            3) Gate | qureg
            4) Gate | (qubit, )
            5) Gate | (qureg, qubit)

        where qubit is a Qubit object and qureg is a Qureg object. This
        function takes the right hand side of | and transforms it to the
        correct input parameter of a Command object which is:

            1) -> Gate | ([qubit], )
            2) -> Gate | ([qubit0, qubit1], )
            3) -> Gate | (qureg, )
            4) -> Gate | ([qubit], )
            5) -> Gate | (qureg, [qubit])

        Args:
            qubits: a Qubit object, a list of Qubit objects, a Qureg object,
                or a tuple of Qubit or Qureg objects (can be mixed).
        Returns:
            Canonical representation (tuple<qureg>): A tuple containing Qureg
            (or list of Qubits) objects.
        """
        if not isinstance(qubits, tuple):
            qubits = (qubits, )

        qubits = list(qubits)

        for i in range(len(qubits)):
            if isinstance(qubits[i], BasicQubit):
                qubits[i] = [qubits[i]]

        return tuple(qubits)

    def generate_command(self, qubits):
        """
        Helper function to generate a command consisting of the gate and
        the qubits being acted upon.

        Args:
            qubits: see BasicGate.make_tuple_of_qureg(qubits)

        Returns:
            A Command object containing the gate and the qubits.
        """
        qubits = self.make_tuple_of_qureg(qubits)

        engines = [q.engine for reg in qubits for q in reg]
        eng = engines[0]
        assert all(e is eng for e in engines)
        return Command(eng, self, qubits)

    def __or__(self, qubits):
        """
        Operator| overload which enables the syntax Gate | qubits.

        Example:
            1) Gate | qubit
            2) Gate | [qubit0, qubit1]
            3) Gate | qureg
            4) Gate | (qubit, )
            5) Gate | (qureg, qubit)

        Args:
            qubits: a Qubit object, a list of Qubit objects, a Qureg object,
                    or a tuple of Qubit or Qureg objects (can be mixed).
        """
        cmd = self.generate_command(qubits)
        apply_command(cmd)

    def __eq__(self, other):
        """
        Equality comparision

        Return True if instance of the same class, unless other is an instance
        of :class:MatrixGate, in which case equality is to be checked by
        testing for existence and (approximate) equality of matrix
        representations.
        """
        if isinstance(other, self.__class__):
            return True
        if isinstance(other, MatrixGate):
            return NotImplemented
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        raise NotImplementedError('This gate does not implement __str__.')

    def to_string(self, symbols):
        """
        String representation

        Achieve same function as str() but can be extended for configurable
        representation
        """
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def is_identity(self):
        return False


class MatrixGate(BasicGate):
    """
    Defines a gate class whose instances are defined by a matrix.

    Note:
        Use this gate class only for gates acting on a small numbers of qubits.
        In general, consider instead using one of the provided ProjectQ gates
        or define a new class as this allows the compiler to work symbolically.

    Example:

        .. code-block:: python

            gate = MatrixGate([[0, 1], [1, 0]])
            gate | qubit
    """
    def __init__(self, matrix=None):
        """
        Initialize MatrixGate

        Args:
            matrix(numpy.matrix): matrix which defines the gate. Default: None
        """
        BasicGate.__init__(self)
        self._matrix = np.matrix(matrix) if matrix is not None else None

    @property
    def matrix(self):
        return self._matrix

    @matrix.setter
    def matrix(self, matrix):
        self._matrix = np.matrix(matrix)

    def __eq__(self, other):
        """
        Equality comparision

        Return True only if both gates have a matrix respresentation and the
        matrices are (approximately) equal. Otherwise return False.
        """
        if not hasattr(other, 'matrix'):
            return False
        if (not isinstance(self.matrix, np.matrix)
                or not isinstance(other.matrix, np.matrix)):
            raise TypeError("One of the gates doesn't have the correct "
                            "type (numpy.matrix) for the matrix "
                            "attribute.")
        if (self.matrix.shape == other.matrix.shape and np.allclose(
                self.matrix, other.matrix, rtol=RTOL, atol=ATOL,
                equal_nan=False)):
            return True
        return False

    def __str__(self):
        return ("MatrixGate(" + str(self.matrix.tolist()) + ")")

    def __hash__(self):
        return hash(str(self))

    def get_inverse(self):
        return MatrixGate(np.linalg.inv(self.matrix))


class SelfInverseGate(BasicGate):
    """
    Self-inverse basic gate class.

    Automatic implementation of the get_inverse-member function for self-
    inverse gates.

    Example:
        .. code-block:: python

            # get_inverse(H) == H, it is a self-inverse gate:
            get_inverse(H) | qubit
    """
    def get_inverse(self):
        return deepcopy(self)


class BasicRotationGate(BasicGate):
    """
    Defines a base class of a rotation gate.

    A rotation gate has a continuous parameter (the angle), labeled 'angle' /
    self.angle. Its inverse is the same gate with the negated argument.
    Rotation gates of the same class can be merged by adding the angles.
    The continuous parameter is modulo 4 * pi, self.angle is in the interval
    [0, 4 * pi).
    """
    def __init__(self, angle):
        """
        Initialize a basic rotation gate.

        Args:
            angle (float): Angle of rotation (saved modulo 4 * pi)
        """
        BasicGate.__init__(self)
        rounded_angle = round(float(angle) % (4. * math.pi), ANGLE_PRECISION)
        if rounded_angle > 4 * math.pi - ANGLE_TOLERANCE:
            rounded_angle = 0.
        self.angle = rounded_angle

    def __str__(self):
        """
        Return the string representation of a BasicRotationGate.

        Returns the class name and the angle as

        .. code-block:: python

            [CLASSNAME]([ANGLE])
        """
        return self.to_string()

    def to_string(self, symbols=False):
        """
        Return the string representation of a BasicRotationGate.

        Args:
            symbols (bool): uses the pi character and round the angle for a
                            more user friendly display if True, full angle
                            written in radian otherwise.
        """
        if symbols:
            angle = ("(" + str(round(self.angle / math.pi, 3))
                     + unicodedata.lookup("GREEK SMALL LETTER PI") + ")")
        else:
            angle = "(" + str(self.angle) + ")"
        return str(self.__class__.__name__) + angle

    def tex_str(self):
        """
        Return the Latex string representation of a BasicRotationGate.

        Returns the class name and the angle as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return (str(self.__class__.__name__) + "$_{"
                + str(round(self.angle / math.pi, 3)) + "\\pi}$")

    def get_inverse(self):
        """
        Return the inverse of this rotation gate (negate the angle, return new
        object).
        """
        if self.angle == 0:
            return self.__class__(0)
        return self.__class__(-self.angle + 4 * math.pi)

    def get_merged(self, other):
        """
        Return self merged with another gate.

        Default implementation handles rotation gate of the same type, where
        angles are simply added.

        Args:
            other: Rotation gate of same type.

        Raises:
            NotMergeable: For non-rotation gates or rotation gates of
                different type.

        Returns:
            New object representing the merged gates.
        """
        if isinstance(other, self.__class__):
            return self.__class__(self.angle + other.angle)
        raise NotMergeable("Can't merge different types of rotation gates.")

    def __eq__(self, other):
        """ Return True if same class and same rotation angle. """
        if isinstance(other, self.__class__):
            return self.angle == other.angle
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))

    def is_identity(self):
        """
        Return True if the gate is equivalent to an Identity gate
        """
        return self.angle == 0. or self.angle == 4 * math.pi


class BasicPhaseGate(BasicGate):
    """
    Defines a base class of a phase gate.

    A phase gate has a continuous parameter (the angle), labeled 'angle' /
    self.angle. Its inverse is the same gate with the negated argument.
    Phase gates of the same class can be merged by adding the angles.
    The continuous parameter is modulo 2 * pi, self.angle is in the interval
    [0, 2 * pi).
    """
    def __init__(self, angle):
        """
        Initialize a basic rotation gate.

        Args:
            angle (float): Angle of rotation (saved modulo 2 * pi)
        """
        BasicGate.__init__(self)
        rounded_angle = round(float(angle) % (2. * math.pi), ANGLE_PRECISION)
        if rounded_angle > 2 * math.pi - ANGLE_TOLERANCE:
            rounded_angle = 0.
        self.angle = rounded_angle

    def __str__(self):
        """
        Return the string representation of a BasicPhaseGate.

        Returns the class name and the angle as

        .. code-block:: python

            [CLASSNAME]([ANGLE])
        """
        return str(self.__class__.__name__) + "(" + str(self.angle) + ")"

    def tex_str(self):
        """
        Return the Latex string representation of a BasicPhaseGate.

        Returns the class name and the angle as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return str(self.__class__.__name__) + "$_{" + str(self.angle) + "}$"

    def get_inverse(self):
        """
        Return the inverse of this rotation gate (negate the angle, return new
        object).
        """
        if self.angle == 0:
            return self.__class__(0)
        return self.__class__(-self.angle + 2 * math.pi)

    def get_merged(self, other):
        """
        Return self merged with another gate.

        Default implementation handles rotation gate of the same type, where
        angles are simply added.

        Args:
            other: Rotation gate of same type.

        Raises:
            NotMergeable: For non-rotation gates or rotation gates of
                different type.

        Returns:
            New object representing the merged gates.
        """
        if isinstance(other, self.__class__):
            return self.__class__(self.angle + other.angle)
        raise NotMergeable("Can't merge different types of rotation gates.")

    def __eq__(self, other):
        """ Return True if same class and same rotation angle. """
        if isinstance(other, self.__class__):
            return self.angle == other.angle
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))


# Classical instruction gates never have control qubits.
class ClassicalInstructionGate(BasicGate):
    """
    Classical instruction gate.

    Base class for all gates which are not quantum gates in the typical sense,
    e.g., measurement, allocation/deallocation, ...
    """
    pass


class FastForwardingGate(ClassicalInstructionGate):
    """
    Base class for classical instruction gates which require a fast-forward
    through compiler engines that cache / buffer gates. Examples include
    Measure and Deallocate, which both should be executed asap, such
    that Measurement results are available and resources are freed,
    respectively.

    Note:
        The only requirement is that FlushGate commands run the entire
        circuit. FastForwardingGate objects can be used but the user cannot
        expect a measurement result to be available for all back-ends when
        calling only Measure. E.g., for the IBM Quantum Experience back-end,
        sending the circuit for each Measure-gate would be too inefficient,
        which is why a final

        .. code-block: python

            eng.flush()

        is required before the circuit gets sent through the API.
    """
    pass


class BasicMathGate(BasicGate):
    """
    Base class for all math gates.

    It allows efficient emulation by providing a mathematical representation
    which is given by the concrete gate which derives from this base class.
    The AddConstant gate, for example, registers a function of the form

    .. code-block:: python

        def add(x):
            return (x+a,)

    upon initialization. More generally, the function takes integers as
    parameters and returns a tuple / list of outputs, each entry corresponding
    to the function input. As an example, consider out-of-place
    multiplication, which takes two input registers and adds the result into a
    third, i.e., (a,b,c) -> (a,b,c+a*b). The corresponding function then is

    .. code-block:: python

        def multiply(a,b,c)
            return (a,b,c+a*b)
    """
    def __init__(self, math_fun):
        """
        Initialize a BasicMathGate by providing the mathematical function that
        it implements.

        Args:
            math_fun (function): Function which takes as many int values as
                input, as the gate takes registers. For each of these values,
                it then returns the output (i.e., it returns a list/tuple of
                output values).

        Example:
            .. code-block:: python

                def add(a,b):
                    return (a,a+b)
                BasicMathGate.__init__(self, add)

        If the gate acts on, e.g., fixed point numbers, the number of bits per
        register is also required in order to describe the action of such a
        mathematical gate. For this reason, there is

        .. code-block:: python

            BasicMathGate.get_math_function(qubits)

        which can be overwritten by the gate deriving from BasicMathGate.

        Example:
            .. code-block:: python

                def get_math_function(self, qubits):
                    n = len(qubits[0])
                    scal = 2.**n
                    def math_fun(a):
                        return (int(scal * (math.sin(math.pi * a / scal))),)
                    return math_fun

        """
        BasicGate.__init__(self)

        def math_function(x):
            return list(math_fun(*x))

        self._math_function = math_function

    def __str__(self):
        return "MATH"

    def get_math_function(self, qubits):
        """
        Return the math function which corresponds to the action of this math
        gate, given the input to the gate (a tuple of quantum registers).

        Args:
            qubits (tuple<Qureg>): Qubits to which the math gate is being
                applied.

        Returns:
            math_fun (function): Python function describing the action of this
            gate. (See BasicMathGate.__init__ for an example).
        """
        return self._math_function



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition rule for global phases.

Deletes global phase gates (which can be ignored).
"""

from projectq.cengines import DecompositionRule
from projectq.meta import get_control_count
from projectq.ops import Ph


def _decompose_PhNoCtrl(cmd):
    """ Throw out global phases (no controls). """
    pass


def _recognize_PhNoCtrl(cmd):
    """ Recognize global phases (no controls). """
    return get_control_count(cmd) == 0


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Ph, _decompose_PhNoCtrl, _recognize_PhNoCtrl)
]



def return_swap_depth(swaps):
    """
    Returns the circuit depth to execute these swaps.

    Args:
        swaps(list of tuples): Each tuple contains two integers representing
                               the two IDs of the qubits involved in the
                               Swap operation
    Returns:
        Circuit depth to execute these swaps.
    """
    depth_of_qubits = dict()
    for qb0_id, qb1_id in swaps:
        if qb0_id not in depth_of_qubits:
            depth_of_qubits[qb0_id] = 0
        if qb1_id not in depth_of_qubits:
            depth_of_qubits[qb1_id] = 0
        max_depth = max(depth_of_qubits[qb0_id], depth_of_qubits[qb1_id])
        depth_of_qubits[qb0_id] = max_depth + 1
        depth_of_qubits[qb1_id] = max_depth + 1
    return max(list(depth_of_qubits.values()) + [0])



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Registers a decomposition for the Ry gate into an Rz and Rx(pi/2) gate.
"""

import math

from projectq.cengines import DecompositionRule
from projectq.meta import Compute, Control, get_control_count, Uncompute
from projectq.ops import Rx, Ry, Rz


def _decompose_ry(cmd):
    """ Decompose the Ry gate."""
    qubit = cmd.qubits[0]
    eng = cmd.engine
    angle = cmd.gate.angle

    with Control(eng, cmd.control_qubits):
        with Compute(eng):
            Rx(math.pi/2.) | qubit
        Rz(angle) | qubit
        Uncompute(eng)


def _recognize_RyNoCtrl(cmd):
    """ For efficiency reasons only if no control qubits."""
    return get_control_count(cmd) == 0


#: Decomposition rules
all_defined_decomposition_rules = [
    DecompositionRule(Ry, _decompose_ry, _recognize_RyNoCtrl)
]



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import inspect
import pkgutil
from importlib import import_module

# Allow extending this namespace.
__path__ = pkgutil.extend_path(__path__, __name__)

from ._core import (BasicEngine, LastEngineException, ForwarderEngine,
                    BasicEngine, LastEngineException, CommandModifier,
                    BasicMapperEngine, MainEngine, NotYetMeasuredError,
                    return_swap_depth)


def dynamic_import(name):
    imported_module = import_module('.' + name, package=__name__)

    for attr_name in dir(imported_module):
        module_attr = getattr(imported_module, attr_name)

        # Only automatically import classes that derive from BasicEngine or
        # Exception and that have not already been imported and avoid
        # importing classes from other ProjectQ submodules
        if (inspect.isclass(module_attr)
                and issubclass(module_attr, (BasicEngine, Exception))
                and attr_name not in globals()
                and __name__ in module_attr.__module__):
            globals()[attr_name] = module_attr

        # If present, import all symbols from the 'all_defined_symbols' list
        if attr_name == 'all_defined_symbols':
            for symbol in module_attr:
                globals()[symbol.__name__] = symbol


_failed_list = []
for (_, pkg_name, _) in pkgutil.iter_modules(path=__path__):
    if pkg_name.endswith('test') or pkg_name == '_core':
        continue
    try:
        dynamic_import(pkg_name)
    except ImportError:
        _failed_list.append(pkg_name)

for pkg_name in _failed_list:
    dynamic_import(pkg_name)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.meta import Dagger


class DecompositionRuleSet:
    """
    A collection of indexed decomposition rules.
    """
    def __init__(self, rules=None, modules=None):
        """
        Args:
            rules list[DecompositionRule]: Initial decomposition rules.
            modules (iterable[ModuleWithDecompositionRuleSet]): A list of
                things with an "all_defined_decomposition_rules" property
                containing decomposition rules to add to the rule set.
        """
        self.decompositions = dict()

        if rules:
            self.add_decomposition_rules(rules)

        if modules:
            self.add_decomposition_rules([
                rule
                for module in modules
                for rule in module.all_defined_decomposition_rules])

    def add_decomposition_rules(self, rules):
        for rule in rules:
            self.add_decomposition_rule(rule)

    def add_decomposition_rule(self, rule):
        """
        Add a decomposition rule to the rule set.

        Args:
            rule (DecompositionRuleGate): The decomposition rule to add.
        """
        decomp_obj = _Decomposition(rule.gate_decomposer, rule.gate_recognizer,
                                    rule.rule_priority)
        cls = rule.gate_class.__name__
        if cls not in self.decompositions:
            self.decompositions[cls] = []
        self.decompositions[cls].append(decomp_obj)


class ModuleWithDecompositionRuleSet:
    """
    Interface type for explaining one of the parameters that can be given to
    DecompositionRuleSet.
    """
    def __init__(self, all_defined_decomposition_rules):
        """
        Args:
            all_defined_decomposition_rules (list[DecompositionRule]):
                A list of decomposition rules.
        """
        self.all_defined_decomposition_rules = all_defined_decomposition_rules


class _Decomposition(object):
    """
    The Decomposition class can be used to register a decomposition rule (by
    calling register_decomposition)
    """
    def __init__(self, replacement_fun, recogn_fun, priority):
        """
        Construct the Decomposition object.

        Args:
            replacement_fun: Function that, when called with a `Command`
                object, decomposes this command.
            recogn_fun: Function that, when called with a `Command` object,
                returns True if and only if the replacement rule can handle
                this command.

        Every Decomposition is registered with the gate class. The
        Decomposition rule is then potentially valid for all objects which are
        an instance of that same class
        (i.e., instance of gate_object.__class__). All other parameters have
        to be checked by the recogn_fun, i.e., it has to decide whether the
        decomposition rule can indeed be applied to replace the given Command.

        As an example, consider recognizing the Toffoli gate, which is a
        Pauli-X gate with 2 control qubits. The recognizer function would then
        be:

        .. code-block:: python

            def recogn_toffoli(cmd):
                # can be applied if the gate is an X-gate with 2 controls:
                return len(cmd.control_qubits) == 2

        and, given a replacement function `replace_toffoli`, the decomposition
        rule can be registered as

        .. code-block:: python

            register_decomposition(X.__class__, decompose_toffoli,
                                   recogn_toffoli)

        Note:
            See projectq.setups.decompositions for more example codes.

        """
        self.decompose = replacement_fun
        self.check = recogn_fun
        self.priority = priority

    def get_inverse_decomposition(self):
        """
        Return the Decomposition object which handles the inverse of the
        original command.

        This simulates the user having added a decomposition rule for the
        inverse as well. Since decomposing the inverse of a command can be
        achieved by running the original decomposition inside a
        `with Dagger(engine):` statement, this is not necessary
        (and will be done automatically by the framework).

        Returns:
            Decomposition handling the inverse of the original command.
        """
        def decomp(cmd):
            with Dagger(cmd.engine):
                self.decompose(cmd.get_inverse())

        def recogn(cmd):
            return self.check(cmd.get_inverse())

        return _Decomposition(decomp, recogn, self.priority)



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from projectq.ops import BasicGate

from ._utils import _exec


class PhaseOracle:
    """
    Synthesizes phase circuit from an arbitrary Boolean function.

    This creates a phase circuit from a Boolean function.  It inverts the phase
    of all amplitudes for which the function evaluates to 1.  The Boolean
    function is provided as integer representation of the function's truth
    table in binary notation.  For example, for the majority-of-three function,
    which truth table 11101000, the value for function can be, e.g.,
    ``0b11101000``, ``0xe8``, or ``232``.

    Note that a phase circuit can only accurately be found for a normal
    function, i.e., a function that maps the input pattern 0, 0, ..., 0 to 0.
    The circuits for a function and its inverse are the same.

    Example:

        This example creates a phase circuit based on the majority-of-three
        function on qubits ``a``, ``b``, and ``c``.

        .. code-block:: python

            PhaseOracle(0x8e) | (a, b, c)
    """

    def __init__(self, function, **kwargs):
        """
        Initializes a phase oracle.

        Args:
            function (int): Function truth table.

        Keyword Args:
            synth: A RevKit synthesis command which creates a reversible
                   circuit based on a truth table and requires no additional
                   ancillae (e.g., ``revkit.esopps``).  Can also be a nullary
                   lambda that calls several RevKit commands.
                   **Default:** ``revkit.esopps``
        """
        if isinstance(function, int):
            self.function = function
        else:
            try:
                import dormouse
                self.function = dormouse.to_truth_table(function)
            except ImportError:  # pragma: no cover
                raise RuntimeError(
                    "The dormouse library needs to be installed in order to "
                    "automatically compile Python code into functions.  Try "
                    "to install dormouse with 'pip install dormouse'."
                )
        self.kwargs = kwargs

        self._check_function()

    def __or__(self, qubits):
        """
        Applies phase circuit to qubits (and synthesizes circuit).

        Args:
            qubits (tuple<Qureg>): Qubits to which the phase circuit is being
                                   applied.
        """
        try:
            import revkit
        except ImportError:  # pragma: no cover
            raise RuntimeError(
                "The RevKit Python library needs to be installed and in the "
                "PYTHONPATH in order to call this function")

        # convert qubits to tuple
        qs = []
        for item in BasicGate.make_tuple_of_qureg(qubits):
            qs += item if isinstance(item, list) else [item]

        # function truth table cannot be larger than number of control qubits
        # allow
        if 2**(2**len(qs)) <= self.function:
            raise AttributeError(
                "Function truth table exceeds number of control qubits")

        # create truth table from function integer
        hex_length = max(2**(len(qs) - 1) // 4, 1)
        revkit.tt(table="{0:#0{1}x}".format(self.function, hex_length))

        # create phase circuit from truth table
        self.kwargs.get("synth", lambda: revkit.esopps())()

        # check whether circuit has correct signature
        if revkit.ps(mct=True, silent=True)['qubits'] != len(qs):
            raise RuntimeError("Generated circuit lines does not match "
                               "provided qubits")

        # convert reversible circuit to ProjectQ code and execute it
        _exec(revkit.to_projectq(mct=True), qs)

    def _check_function(self):
        """
        Checks whether function is valid.
        """
        # function must be positive.  We check in __or__ whether function is
        # too large
        if self.function < 0:
            raise AttributeError("Function must be a postive integer")



#   Copyright 2017 ProjectQ-Framework (www.projectq.ch)
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Contains meta gates, i.e.,
* DaggeredGate (Represents the inverse of an arbitrary gate)
* ControlledGate (Represents a controlled version of an arbitrary gate)
* Tensor/All (Applies a single qubit gate to all supplied qubits), e.g.,
    Example:
        .. code-block:: python

          Tensor(H) | (qubit1, qubit2) # apply H to qubit #1 and #2

As well as the meta functions
* get_inverse (Tries to access the get_inverse member function of a gate
               and upon failure returns a DaggeredGate)
* C (Creates an n-ary controlled version of an arbitrary gate)
"""

from ._basics import BasicGate, NotInvertible


class ControlQubitError(Exception):
    """
    Exception thrown when wrong number of control qubits are supplied.
    """
    pass


class DaggeredGate(BasicGate):
    """
    Wrapper class allowing to execute the inverse of a gate, even when it does
    not define one.

    If there is a replacement available, then there is also one for the
    inverse, namely the replacement function run in reverse, while inverting
    all gates. This class enables using this emulation automatically.

    A DaggeredGate is returned automatically when employing the get_inverse-
    function on a gate which does not provide a get_inverse() member function.

    Example:
        .. code-block:: python

            with Dagger(eng):
                MySpecialGate | qubits

    will create a DaggeredGate if MySpecialGate does not implement
    get_inverse. If there is a decomposition function available, an auto-
    replacer engine can automatically replace the inverted gate by a call to
    the decomposition function inside a "with Dagger"-statement.
    """

    def __init__(self, gate):
        """
        Initialize a DaggeredGate representing the inverse of the gate 'gate'.

        Args:
            gate: Any gate object of which to represent the inverse.
        """
        BasicGate.__init__(self)
        self._gate = gate

        try:
            # Hermitian conjugate is inverse matrix
            self.matrix = gate.matrix.getH()
        except AttributeError:
            pass

    def __str__(self):
        r"""
        Return string representation (str(gate) + \"^\dagger\").
        """
        return str(self._gate) + r"^\dagger"

    def tex_str(self):
        """
        Return the Latex string representation of a Daggered gate.
        """
        if hasattr(self._gate, 'tex_str'):
            return self._gate.tex_str() + r"${}^\dagger$"
        return str(self._gate) + r"${}^\dagger$"

    def get_inverse(self):
        """
        Return the inverse gate (the inverse of the inverse of a gate is the
        gate itself).
        """
        return self._gate

    def __eq__(self, other):
        """
        Return True if self is equal to other, i.e., same type and
        representing the inverse of the same gate.
        """
        return isinstance(other, self.__class__) and self._gate == other._gate

    def __hash__(self):
        return hash(str(self))


def get_inverse(gate):
    """
    Return the inverse of a gate.

    Tries to call gate.get_inverse and, upon failure, creates a DaggeredGate
    instead.

    Args:
        gate: Gate of which to get the inverse

    Example:
        .. code-block:: python

            get_inverse(H) # returns a Hadamard gate (HGate object)
    """
    try:
        return gate.get_inverse()
    except NotInvertible:
        return DaggeredGate(gate)


def is_identity(gate):
    """
    Return True if the gate is an identity gate.

    Tries to call gate.is_identity and, upon failure, returns False

    Args:
        gate: Gate of which to get the inverse

    Example:
        .. code-block:: python

            get_inverse(Rx(2*math.pi)) # returns True
            get_inverse(Rx(math.pi)) # returns False
    """
    return gate.is_identity()


class ControlledGate(BasicGate):
    """
    Controlled version of a gate.

    Note:
        Use the meta function :func:`C()` to create a controlled gate

    A wrapper class which enables (multi-) controlled gates. It overloads
    the __or__-operator, using the first qubits provided as control qubits.
    The n control-qubits need to be the first n qubits. They can be in
    separate quregs.

    Example:
        .. code-block:: python

            ControlledGate(gate, 2) | (qb0, qb2, qb3) # qb0 & qb2 are controls
            C(gate, 2) | (qb0, qb2, qb3) # This is much nicer.
            C(gate, 2) | ([qb0,qb2], qb3) # Is equivalent

    Note:
        Use :func:`C` rather than ControlledGate, i.e.,

        .. code-block:: python

            C(X, 2) == Toffoli
    """

    def __init__(self, gate, n=1):
        """
        Initialize a ControlledGate object.

        Args:
            gate: Gate to wrap.
            n (int): Number of control qubits.
        """
        BasicGate.__init__(self)
        if isinstance(gate, ControlledGate):
            self._gate = gate._gate
            self._n = gate._n + n
        else:
            self._gate = gate
            self._n = n

    def __str__(self):
        """ Return string representation, i.e., CC...C(gate). """
        return "C" * self._n + str(self._gate)

    def get_inverse(self):
        """
        Return inverse of a controlled gate, which is the controlled inverse
        gate.
        """
        return ControlledGate(get_inverse(self._gate), self._n)

    def __or__(self, qubits):
        """
        Apply the controlled gate to qubits, using the first n qubits as
        controls.

        Note: The control qubits can be split across the first quregs.
            However, the n-th control qubit needs to be the last qubit in a
            qureg. The following quregs belong to the gate.

        Args:
            qubits (tuple of lists of Qubit objects): qubits to which to apply
                the gate.
        """
        qubits = BasicGate.make_tuple_of_qureg(qubits)

        ctrl = []
        gate_quregs = []
        adding_to_controls = True
        for reg in qubits:
            if adding_to_controls:
                ctrl += reg
                adding_to_controls = len(ctrl) < self._n
            else:
                gate_quregs.append(reg)
        # Test that there were enough control quregs and that that
        # the last control qubit was the last qubit in a qureg.
        if len(ctrl) != self._n:
            raise ControlQubitError("Wrong number of control qubits. "
                                    "First qureg(s) need to contain exactly "
                                    "the required number of control quregs.")

        import projectq.meta
        with projectq.meta.Control(gate_quregs[0][0].engine, ctrl):
            self._gate | tuple(gate_quregs)

    def __eq__(self, other):
        """ Compare two ControlledGate objects (return True if equal). """
        return (isinstance(other, self.__class__) and self._gate == other._gate
                and self._n == other._n)

    def __ne__(self, other):
        return not self.__eq__(other)


def C(gate, n=1):
    """
    Return n-controlled version of the provided gate.

    Args:
        gate: Gate to turn into its controlled version
        n: Number of controls (default: 1)

    Example:
        .. code-block:: python

            C(NOT) | (c, q) # equivalent to CNOT | (c, q)
    """
    return ControlledGate(gate, n)


class Tensor(BasicGate):
    """
    Wrapper class allowing to apply a (single-qubit) gate to every qubit in a
    quantum register. Allowed syntax is to supply either a qureg or a tuple
    which contains only one qureg.

    Example:
        .. code-block:: python

            Tensor(H) | x # applies H to every qubit in the list of qubits x
            Tensor(H) | (x,) # alternative to be consistent with other syntax
    """

    def __init__(self, gate):
        """ Initialize a Tensor object for the gate. """
        BasicGate.__init__(self)
        self._gate = gate

    def __str__(self):
        """ Return string representation. """
        return "Tensor(" + str(self._gate) + ")"

    def get_inverse(self):
        """
        Return the inverse of this tensored gate (which is the tensored
        inverse of the gate).
        """
        return Tensor(get_inverse(self._gate))

    def __eq__(self, other):
        return isinstance(other, Tensor) and self._gate == other._gate

    def __ne__(self, other):
        return not self.__eq__(other)

    def __or__(self, qubits):
        """ Applies the gate to every qubit in the quantum register qubits. """
        if isinstance(qubits, tuple):
            assert len(qubits) == 1
            qubits = qubits[0]
        assert isinstance(qubits, list)
        for qubit in qubits:
            self._gate | qubit


#: Shortcut (instance of) :class:`projectq.ops.Tensor`
All = Tensor


all_defined_symbols = [get_inverse, C, is_identity]
